@@ -61,6 +61,9 @@ extern "C" {
 #define G_SUPPLICANT_KEYMGMT_WPA_EAP_256	(1 << 8)
 #define G_SUPPLICANT_KEYMGMT_WPS		(1 << 9)
 
+#define G_SUPPLICANT_PROTO_WPA		(1 << 0)
+#define G_SUPPLICANT_PROTO_RSN		(1 << 1)
+
 #define G_SUPPLICANT_GROUP_WEP40		(1 << 0)
 #define G_SUPPLICANT_GROUP_WEP104		(1 << 1)
 #define G_SUPPLICANT_GROUP_TKIP		(1 << 2)
@@ -70,10 +73,13 @@ extern "C" {
 #define G_SUPPLICANT_PAIRWISE_TKIP	(1 << 1)
 #define G_SUPPLICANT_PAIRWISE_CCMP	(1 << 2)
 
+#define G_SUPPLICANT_MAX_FAST_SCAN	4
+
 typedef enum {
 	G_SUPPLICANT_MODE_UNKNOWN,
 	G_SUPPLICANT_MODE_INFRA,
 	G_SUPPLICANT_MODE_IBSS,
+	G_SUPPLICANT_MODE_MASTER,
 } GSupplicantMode;
 
 typedef enum {
@@ -106,8 +112,13 @@ typedef enum {
 struct _GSupplicantSSID {
 	const void *ssid;
 	unsigned int ssid_len;
+	unsigned int scan_ssid;
 	GSupplicantMode mode;
 	GSupplicantSecurity security;
+	unsigned int protocol;
+	unsigned int pairwise_cipher;
+	unsigned int group_cipher;
+	unsigned int freq;
 	const char *eap;
 	const char *passphrase;
 	const char *identity;
@@ -118,10 +129,24 @@ struct _GSupplicantSSID {
 	const char *phase2_auth;
 	dbus_bool_t use_wps;
 	const char *pin_wps;
+	const char *bgscan;
 };
 
 typedef struct _GSupplicantSSID GSupplicantSSID;
 
+struct _GSupplicantScanParams {
+	struct scan_ssid {
+		unsigned char ssid[32];
+		uint8_t ssid_len;
+	} ssids[G_SUPPLICANT_MAX_FAST_SCAN];
+
+	uint8_t num_ssids;
+
+	uint16_t freqs[G_SUPPLICANT_MAX_FAST_SCAN];
+};
+
+typedef struct _GSupplicantScanParams GSupplicantScanParams;
+
 /* global API */
 typedef void (*GSupplicantCountryCallback) (void *user_data);
 
@@ -139,12 +164,14 @@ typedef void (*GSupplicantInterfaceCallback) (int result,
 							void *user_data);
 
 int g_supplicant_interface_create(const char *ifname, const char *driver,
+					const char *bridge,
 					GSupplicantInterfaceCallback callback,
 							void *user_data);
 int g_supplicant_interface_remove(GSupplicantInterface *interface,
 					GSupplicantInterfaceCallback callback,
 							void *user_data);
 int g_supplicant_interface_scan(GSupplicantInterface *interface,
+					GSupplicantScanParams *scan_data,
 					GSupplicantInterfaceCallback callback,
 							void *user_data);
 
@@ -157,6 +184,9 @@ int g_supplicant_interface_disconnect(GSupplicantInterface *interface,
 					GSupplicantInterfaceCallback callback,
 							void *user_data);
 
+int g_supplicant_interface_set_apscan(GSupplicantInterface *interface,
+							unsigned int ap_scan);
+
 void g_supplicant_interface_set_data(GSupplicantInterface *interface,
 								void *data);
 void *g_supplicant_interface_get_data(GSupplicantInterface *interface);
@@ -167,6 +197,13 @@ const char *g_supplicant_interface_get_wps_key(GSupplicantInterface *interface);
 const void *g_supplicant_interface_get_wps_ssid(GSupplicantInterface *interface,
 							unsigned int *ssid_len);
 GSupplicantWpsState g_supplicant_interface_get_wps_state(GSupplicantInterface *interface);
+unsigned int g_supplicant_interface_get_mode(GSupplicantInterface *interface);
+dbus_bool_t g_supplicant_interface_get_ready(GSupplicantInterface *interface);
+unsigned int g_supplicant_interface_get_max_scan_ssids(
+					GSupplicantInterface *interface);
+
+int g_supplicant_interface_enable_selected_network(GSupplicantInterface *interface,
+							dbus_bool_t enable);
 
 /* Network API */
 struct _GSupplicantNetwork;
@@ -182,6 +219,7 @@ const void *g_supplicant_network_get_ssid(GSupplicantNetwork *network,
 const char *g_supplicant_network_get_mode(GSupplicantNetwork *network);
 const char *g_supplicant_network_get_security(GSupplicantNetwork *network);
 dbus_int16_t g_supplicant_network_get_signal(GSupplicantNetwork *network);
+dbus_uint16_t g_supplicant_network_get_frequency(GSupplicantNetwork *network);
 dbus_bool_t g_supplicant_network_get_wps(GSupplicantNetwork *network);
 
 struct _GSupplicantCallbacks {
@@ -194,6 +232,8 @@ struct _GSupplicantCallbacks {
 	void (*scan_finished) (GSupplicantInterface *interface);
 	void (*network_added) (GSupplicantNetwork *network);
 	void (*network_removed) (GSupplicantNetwork *network);
+	void (*network_changed) (GSupplicantNetwork *network,
+					const char *property);
 	void (*debug) (const char *str);
 };
 
