@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -33,6 +33,7 @@
 #include <glib.h>
 
 #define CONNMAN_API_SUBJECT_TO_CHANGE
+#include <connman/technology.h>
 #include <connman/plugin.h>
 #include <connman/device.h>
 #include <connman/inet.h>
@@ -81,7 +82,8 @@ static struct connman_network_driver cable_driver = {
 	.disconnect	= cable_disconnect,
 };
 
-static void add_network(struct connman_device *device)
+static void add_network(struct connman_device *device,
+			struct ethernet_data *ethernet)
 {
 	struct connman_network *network;
 	int index;
@@ -95,7 +97,6 @@ static void add_network(struct connman_device *device)
 	connman_network_set_index(network, index);
 
 	connman_network_set_name(network, "Wired");
-	connman_network_set_protocol(network, CONNMAN_NETWORK_PROTOCOL_IP);
 
 	if (connman_device_add_network(device, network) < 0) {
 		connman_network_unref(network);
@@ -106,9 +107,21 @@ static void add_network(struct connman_device *device)
 
 	connman_network_set_group(network, "cable");
 
-	connman_network_set_method(network, CONNMAN_IPCONFIG_METHOD_DHCP);
-
 	connman_network_set_connected(network, TRUE);
+
+	ethernet->network = network;
+}
+
+static void remove_network(struct connman_device *device,
+				struct ethernet_data *ethernet)
+{
+	if (ethernet->network == NULL)
+		return;
+
+	connman_device_remove_network(device, ethernet->network);
+	connman_network_unref(ethernet->network);
+
+	ethernet->network = NULL;
 }
 
 static void ethernet_newlink(unsigned flags, unsigned change, void *user_data)
@@ -131,10 +144,10 @@ static void ethernet_newlink(unsigned flags, unsigned change, void *user_data)
 	if ((ethernet->flags & IFF_LOWER_UP) != (flags & IFF_LOWER_UP)) {
 		if (flags & IFF_LOWER_UP) {
 			DBG("carrier on");
-			add_network(device);
+			add_network(device, ethernet);
 		} else {
 			DBG("carrier off");
-			connman_device_remove_all_networks(device);
+			remove_network(device, ethernet);
 		}
 	}
 
@@ -172,7 +185,7 @@ static void ethernet_remove(struct connman_device *device)
 
 	connman_rtnl_remove_watch(ethernet->watch);
 
-	connman_device_remove_all_networks(device);
+	remove_network(device, ethernet);
 
 	g_free(ethernet);
 }
@@ -204,10 +217,123 @@ static struct connman_device_driver ethernet_driver = {
 	.disable	= ethernet_disable,
 };
 
+static GList *cdc_interface_list = NULL;
+
+static void tech_add_interface(struct connman_technology *technology,
+			int index, const char *name, const char *ident)
+{
+	DBG("index %d name %s ident %s", index, name, ident);
+
+	if (g_list_find(cdc_interface_list,
+			GINT_TO_POINTER((int) index)) != NULL)
+		return;
+
+	cdc_interface_list = g_list_prepend(cdc_interface_list,
+					(GINT_TO_POINTER((int) index)));
+}
+
+static void tech_remove_interface(struct connman_technology *technology,
+								int index)
+{
+	DBG("index %d", index);
+
+	cdc_interface_list = g_list_remove(cdc_interface_list,
+					GINT_TO_POINTER((int) index));
+}
+
+static void enable_tethering(struct connman_technology *technology,
+						const char *bridge)
+{
+	GList *list;
+
+	for (list = cdc_interface_list; list; list = list->next) {
+		int index = GPOINTER_TO_INT(list->data);
+
+		connman_technology_tethering_notify(technology, TRUE);
+
+		connman_inet_ifup(index);
+
+		connman_inet_add_to_bridge(index, bridge);
+	}
+}
+
+static void disable_tethering(struct connman_technology *technology,
+						const char *bridge)
+{
+	GList *list;
+
+	for (list = cdc_interface_list; list; list = list->next) {
+		int index = GPOINTER_TO_INT(list->data);
+
+		connman_inet_remove_from_bridge(index, bridge);
+
+		connman_inet_ifdown(index);
+
+		connman_technology_tethering_notify(technology, FALSE);
+	}
+}
+
+static int tech_set_tethering(struct connman_technology *technology,
+				const char *identifier, const char *passphrase,
+				const char *bridge, connman_bool_t enabled)
+{
+	DBG("bridge %s enabled %d", bridge, enabled);
+
+	if (enabled)
+		enable_tethering(technology, bridge);
+	else
+		disable_tethering(technology, bridge);
+
+	return 0;
+}
+
+static int tech_probe(struct connman_technology *technology)
+{
+	return 0;
+}
+
+static void tech_remove(struct connman_technology *technology)
+{
+	g_list_free(cdc_interface_list);
+
+	cdc_interface_list = NULL;
+}
+
+static struct connman_technology_driver tech_driver = {
+	.name			= "cdc_ethernet",
+	.type			= CONNMAN_SERVICE_TYPE_GADGET,
+	.probe			= tech_probe,
+	.remove			= tech_remove,
+	.add_interface		= tech_add_interface,
+	.remove_interface 	= tech_remove_interface,
+	.set_tethering		= tech_set_tethering,
+};
+
+static int eth_probe(struct connman_technology *technology)
+{
+	return 0;
+}
+
+static void eth_remove(struct connman_technology *technology)
+{
+	DBG("");
+}
+
+static struct connman_technology_driver eth_driver = {
+	.name			= "ethernet",
+	.type			= CONNMAN_SERVICE_TYPE_ETHERNET,
+	.probe			= eth_probe,
+	.remove			= eth_remove,
+};
+
 static int ethernet_init(void)
 {
 	int err;
 
+	err = connman_technology_driver_register(&eth_driver);
+	if (err < 0)
+		return err;
+
 	err = connman_network_driver_register(&cable_driver);
 	if (err < 0)
 		return err;
@@ -218,11 +344,22 @@ static int ethernet_init(void)
 		return err;
 	}
 
+	err = connman_technology_driver_register(&tech_driver);
+	if (err < 0) {
+		connman_device_driver_unregister(&ethernet_driver);
+		connman_network_driver_unregister(&cable_driver);
+		return err;
+	}
+
 	return 0;
 }
 
 static void ethernet_exit(void)
 {
+	connman_technology_driver_unregister(&eth_driver);
+
+	connman_technology_driver_unregister(&tech_driver);
+
 	connman_network_driver_unregister(&cable_driver);
 
 	connman_device_driver_unregister(&ethernet_driver);
