@@ -26,6 +26,7 @@
 #include <unistd.h>
 #include <stdlib.h>
 #include <errno.h>
+#include <stdio.h>
 #include <string.h>
 #include <sys/ioctl.h>
 #include <sys/socket.h>
@@ -48,19 +49,31 @@
 #include <connman/technology.h>
 #include <connman/log.h>
 #include <connman/option.h>
+#include <connman/storage.h>
 
 #include <gsupplicant/gsupplicant.h>
 
 #define CLEANUP_TIMEOUT   8	/* in seconds */
 #define INACTIVE_TIMEOUT  12	/* in seconds */
+#define MAXIMUM_RETRIES   4
+
+#define BGSCAN_DEFAULT "simple:30:-45:300"
 
 struct connman_technology *wifi_technology = NULL;
 
+struct hidden_params {
+	char ssid[32];
+	unsigned int ssid_len;
+	char *identity;
+	char *passphrase;
+};
+
 struct wifi_data {
 	char *identifier;
 	struct connman_device *device;
 	struct connman_network *network;
 	struct connman_network *pending_network;
+	GSList *networks;
 	GSupplicantInterface *interface;
 	GSupplicantState state;
 	connman_bool_t connected;
@@ -71,6 +84,8 @@ struct wifi_data {
 	int index;
 	unsigned flags;
 	unsigned int watch;
+	int retries;
+	struct hidden_params *hidden;
 };
 
 static GList *iface_list = NULL;
@@ -154,22 +169,36 @@ static int wifi_probe(struct connman_device *device)
 	return 0;
 }
 
+static void remove_networks(struct connman_device *device,
+				struct wifi_data *wifi)
+{
+	GSList *list;
+
+	for (list = wifi->networks; list != NULL; list = list->next) {
+		struct connman_network *network = list->data;
+
+		connman_device_remove_network(device, network);
+		connman_network_unref(network);
+	}
+
+	g_slist_free(wifi->networks);
+	wifi->networks = NULL;
+}
+
 static void wifi_remove(struct connman_device *device)
 {
 	struct wifi_data *wifi = connman_device_get_data(device);
 
-	DBG("device %p", device);
+	DBG("device %p wifi %p", device, wifi);
 
 	if (wifi == NULL)
 		return;
 
 	iface_list = g_list_remove(iface_list, wifi);
 
-	if (wifi->pending_network != NULL) {
-		connman_network_unref(wifi->pending_network);
-		wifi->pending_network = NULL;
-	}
+	remove_networks(device, wifi);
 
+	connman_device_set_powered(device, FALSE);
 	connman_device_set_data(device, NULL);
 	connman_device_unref(wifi->device);
 	connman_rtnl_remove_watch(wifi->watch);
@@ -186,74 +215,266 @@ static void interface_create_callback(int result,
 {
 	struct wifi_data *wifi = user_data;
 
-	DBG("result %d ifname %s", result,
-				g_supplicant_interface_get_ifname(interface));
+	DBG("result %d ifname %s, wifi %p", result,
+				g_supplicant_interface_get_ifname(interface),
+				wifi);
 
-	if (result < 0)
+	if (result < 0 || wifi == NULL)
 		return;
 
 	wifi->interface = interface;
 	g_supplicant_interface_set_data(interface, wifi);
-}
 
-static void interface_remove_callback(int result,
-					GSupplicantInterface *interface,
-							void *user_data)
-{
-	struct wifi_data *wifi = user_data;
-
-	DBG("result %d", result);
-
-	if (result < 0)
+	if (g_supplicant_interface_get_ready(interface) == FALSE)
 		return;
 
-	wifi->interface = NULL;
-}
+	DBG("interface is ready wifi %p tethering %d", wifi, wifi->tethering);
 
+	if (wifi->device == NULL) {
+		connman_error("WiFi device not set");
+		return;
+	}
+
+	connman_device_set_powered(wifi->device, TRUE);
+}
 
 static int wifi_enable(struct connman_device *device)
 {
 	struct wifi_data *wifi = connman_device_get_data(device);
 	const char *interface = connman_device_get_string(device, "Interface");
 	const char *driver = connman_option_get_string("wifi");
+	int ret;
 
 	DBG("device %p %p", device, wifi);
 
-	return g_supplicant_interface_create(interface, driver, NULL,
+	ret = g_supplicant_interface_create(interface, driver, NULL,
 						interface_create_callback,
 							wifi);
+	if (ret < 0)
+		return ret;
+
+	return -EINPROGRESS;
 }
 
 static int wifi_disable(struct connman_device *device)
 {
 	struct wifi_data *wifi = connman_device_get_data(device);
+	int ret;
 
 	DBG("device %p", device);
 
 	wifi->connected = FALSE;
 	wifi->disconnecting = FALSE;
 
-	if (wifi->pending_network != NULL) {
-		connman_network_unref(wifi->pending_network);
+	if (wifi->pending_network != NULL)
 		wifi->pending_network = NULL;
-	}
 
-	return g_supplicant_interface_remove(wifi->interface,
-						interface_remove_callback,
-							wifi);
+	remove_networks(device, wifi);
+
+	ret = g_supplicant_interface_remove(wifi->interface, NULL, NULL);
+	if (ret < 0)
+		return ret;
+
+	return -EINPROGRESS;
+}
+
+static void hidden_free(struct hidden_params *hidden)
+{
+	if (hidden == NULL)
+		return;
+
+	g_free(hidden->identity);
+	g_free(hidden->passphrase);
+	g_free(hidden);
 }
 
 static void scan_callback(int result, GSupplicantInterface *interface,
 						void *user_data)
 {
 	struct connman_device *device = user_data;
+	struct wifi_data *wifi = connman_device_get_data(device);
 
 	DBG("result %d", result);
 
+	if (wifi != NULL && wifi->hidden != NULL) {
+		hidden_free(wifi->hidden);
+		wifi->hidden = NULL;
+	}
+
 	if (result < 0)
 		connman_device_reset_scanning(device);
-	else
-		connman_device_set_scanning(device, FALSE);
+
+	connman_device_set_scanning(device, FALSE);
+	connman_device_unref(device);
+}
+
+static int add_scan_param(gchar *hex_ssid, int freq,
+			GSupplicantScanParams *scan_data,
+			int driver_max_scan_ssids)
+{
+	unsigned int i;
+
+	if (driver_max_scan_ssids > scan_data->num_ssids && hex_ssid != NULL) {
+		gchar *ssid;
+		unsigned int j = 0, hex;
+		size_t hex_ssid_len = strlen(hex_ssid);
+
+		ssid = g_try_malloc0(hex_ssid_len / 2);
+		if (ssid == NULL)
+			return -ENOMEM;
+
+		for (i = 0; i < hex_ssid_len; i += 2) {
+			sscanf(hex_ssid + i, "%02x", &hex);
+			ssid[j++] = hex;
+		}
+
+		memcpy(scan_data->ssids[scan_data->num_ssids].ssid, ssid, j);
+		scan_data->ssids[scan_data->num_ssids].ssid_len = j;
+		scan_data->num_ssids++;
+
+		g_free(ssid);
+	}
+
+	/* Don't add duplicate entries */
+	for (i = 0; i < G_SUPPLICANT_MAX_FAST_SCAN; i++) {
+		if (scan_data->freqs[i] == 0) {
+			scan_data->freqs[i] = freq;
+			break;
+		} else if (scan_data->freqs[i] == freq)
+			break;
+	}
+
+	return 0;
+}
+
+struct last_connected {
+	GTimeVal modified;
+	gchar *ssid;
+	int freq;
+};
+
+static gint sort_entry(gconstpointer a, gconstpointer b, gpointer user_data)
+{
+	GTimeVal *aval = (GTimeVal *)a;
+	GTimeVal *bval = (GTimeVal *)b;
+
+	/* Note that the sort order is descending */
+	if (aval->tv_sec < bval->tv_sec)
+		return 1;
+
+	if (aval->tv_sec > bval->tv_sec)
+		return -1;
+
+	return 0;
+}
+
+static void free_entry(gpointer data)
+{
+	struct last_connected *entry = data;
+
+	g_free(entry->ssid);
+	g_free(entry);
+}
+
+static int get_latest_connections(int max_ssids,
+				GSupplicantScanParams *scan_data)
+{
+	GSequenceIter *iter;
+	GSequence *latest_list;
+	struct last_connected *entry;
+	GKeyFile *keyfile;
+	GTimeVal modified;
+	gchar **services;
+	gchar *str;
+	char *ssid;
+	int i, freq;
+	int num_ssids = 0;
+
+	latest_list = g_sequence_new(free_entry);
+	if (latest_list == NULL)
+		return -ENOMEM;
+
+	services = connman_storage_get_services();
+	for (i = 0; services && services[i]; i++) {
+		if (strncmp(services[i], "wifi_", 5) != 0)
+			continue;
+
+		keyfile = connman_storage_load_service(services[i]);
+
+		str = g_key_file_get_string(keyfile,
+					services[i], "Favorite", NULL);
+		if (str == NULL || g_strcmp0(str, "true")) {
+			if (str)
+				g_free(str);
+			g_key_file_free(keyfile);
+			continue;
+		}
+		g_free(str);
+
+		str = g_key_file_get_string(keyfile,
+					services[i], "AutoConnect", NULL);
+		if (str == NULL || g_strcmp0(str, "true")) {
+			if (str)
+				g_free(str);
+			g_key_file_free(keyfile);
+			continue;
+		}
+		g_free(str);
+
+		str = g_key_file_get_string(keyfile,
+					services[i], "Modified", NULL);
+		if (str != NULL) {
+			g_time_val_from_iso8601(str, &modified);
+			g_free(str);
+		}
+
+		ssid = g_key_file_get_string(keyfile,
+					services[i], "SSID", NULL);
+
+		freq = g_key_file_get_integer(keyfile, services[i],
+					"Frequency", NULL);
+		if (freq) {
+			entry = g_try_new(struct last_connected, 1);
+			if (entry == NULL) {
+				g_sequence_free(latest_list);
+				g_key_file_free(keyfile);
+				g_free(ssid);
+				return -ENOMEM;
+			}
+
+			entry->ssid = ssid;
+			entry->modified = modified;
+			entry->freq = freq;
+
+			g_sequence_insert_sorted(latest_list, entry,
+						sort_entry, NULL);
+			num_ssids++;
+		} else
+			g_free(ssid);
+
+		g_key_file_free(keyfile);
+	}
+
+	g_strfreev(services);
+
+	num_ssids = num_ssids > G_SUPPLICANT_MAX_FAST_SCAN ?
+		G_SUPPLICANT_MAX_FAST_SCAN : num_ssids;
+
+	iter = g_sequence_get_begin_iter(latest_list);
+
+	for (i = 0; i < num_ssids; i++) {
+		entry = g_sequence_get(iter);
+
+		DBG("ssid %s freq %d modified %lu", entry->ssid, entry->freq,
+						entry->modified.tv_sec);
+
+		add_scan_param(entry->ssid, entry->freq, scan_data, max_ssids);
+
+		iter = g_sequence_iter_next(iter);
+	}
+
+	g_sequence_free(latest_list);
+	return num_ssids;
 }
 
 static int wifi_scan(struct connman_device *device)
@@ -266,10 +487,104 @@ static int wifi_scan(struct connman_device *device)
 	if (wifi->tethering == TRUE)
 		return 0;
 
-	ret = g_supplicant_interface_scan(wifi->interface, scan_callback,
-								device);
+	connman_device_ref(device);
+	ret = g_supplicant_interface_scan(wifi->interface, NULL,
+					scan_callback, device);
 	if (ret == 0)
 		connman_device_set_scanning(device, TRUE);
+	else
+		connman_device_unref(device);
+
+	return ret;
+}
+
+static int wifi_scan_fast(struct connman_device *device)
+{
+	struct wifi_data *wifi = connman_device_get_data(device);
+	GSupplicantScanParams *scan_params = NULL;
+	int ret;
+	int driver_max_ssids = 0;
+
+	DBG("device %p %p", device, wifi->interface);
+
+	if (wifi->tethering == TRUE)
+		return 0;
+
+	driver_max_ssids = g_supplicant_interface_get_max_scan_ssids(
+							wifi->interface);
+	DBG("max ssids %d", driver_max_ssids);
+	if (driver_max_ssids == 0)
+		return wifi_scan(device);
+
+	scan_params = g_try_malloc0(sizeof(GSupplicantScanParams));
+	if (scan_params == NULL)
+		return -ENOMEM;
+
+	ret = get_latest_connections(driver_max_ssids, scan_params);
+	if (ret <= 0) {
+		g_free(scan_params);
+		return wifi_scan(device);
+	}
+
+	connman_device_ref(device);
+	ret = g_supplicant_interface_scan(wifi->interface, scan_params,
+						scan_callback, device);
+	if (ret == 0)
+		connman_device_set_scanning(device, TRUE);
+	else {
+		g_free(scan_params);
+		connman_device_unref(device);
+	}
+
+	return ret;
+}
+
+static int wifi_scan_hidden(struct connman_device *device,
+		const char *ssid, unsigned int ssid_len,
+		const char *identity, const char* passphrase)
+{
+	struct wifi_data *wifi = connman_device_get_data(device);
+	GSupplicantScanParams *scan_params = NULL;
+	struct hidden_params *hidden;
+	int ret;
+
+	DBG("hidden SSID %s", ssid);
+
+	if (wifi->tethering == TRUE || wifi->hidden != NULL)
+		return -EBUSY;
+
+	if (ssid == NULL || ssid_len == 0 || ssid_len > 32)
+		return -EINVAL;
+
+	scan_params = g_try_malloc0(sizeof(GSupplicantScanParams));
+	if (scan_params == NULL)
+		return -ENOMEM;
+	memcpy(scan_params->ssids[0].ssid, ssid, ssid_len);
+	scan_params->ssids[0].ssid_len = ssid_len;
+	scan_params->num_ssids = 1;
+
+	hidden = g_try_new0(struct hidden_params, 1);
+	if (hidden == NULL) {
+		g_free(scan_params);
+		return -ENOMEM;
+	}
+	memcpy(hidden->ssid, ssid, ssid_len);
+	hidden->ssid_len = ssid_len;
+	hidden->identity = g_strdup(identity);
+	hidden->passphrase = g_strdup(passphrase);
+	wifi->hidden = hidden;
+
+	connman_device_ref(device);
+	ret = g_supplicant_interface_scan(wifi->interface, scan_params,
+			scan_callback, device);
+	if (ret == 0)
+		connman_device_set_scanning(device, TRUE);
+	else {
+		connman_device_unref(device);
+		g_free(scan_params);
+		hidden_free(wifi->hidden);
+		wifi->hidden = NULL;
+	}
 
 	return ret;
 }
@@ -283,6 +598,8 @@ static struct connman_device_driver wifi_ng_driver = {
 	.enable		= wifi_enable,
 	.disable	= wifi_disable,
 	.scan		= wifi_scan,
+	.scan_fast	= wifi_scan_fast,
+	.scan_hidden    = wifi_scan_hidden,
 };
 
 static void system_ready(void)
@@ -309,7 +626,19 @@ static int network_probe(struct connman_network *network)
 
 static void network_remove(struct connman_network *network)
 {
+	struct connman_device *device = connman_network_get_device(network);
+	struct wifi_data *wifi;
+
 	DBG("network %p", network);
+
+	wifi = connman_device_get_data(device);
+	if (wifi == NULL)
+		return;
+
+	if (wifi->network != network)
+		return;
+
+	wifi->network = NULL;
 }
 
 static void connect_callback(int result, GSupplicantInterface *interface,
@@ -348,7 +677,7 @@ static GSupplicantSecurity network_security(const char *security)
 
 static void ssid_init(GSupplicantSSID *ssid, struct connman_network *network)
 {
-	const char *security, *passphrase;
+	const char *security, *passphrase, *agent_passphrase;
 
 	memset(ssid, 0, sizeof(*ssid));
 	ssid->mode = G_SUPPLICANT_MODE_INFRA;
@@ -359,9 +688,17 @@ static void ssid_init(GSupplicantSSID *ssid, struct connman_network *network)
 	ssid->security = network_security(security);
 	passphrase = connman_network_get_string(network,
 						"WiFi.Passphrase");
-	if (passphrase == NULL || strlen(passphrase) == 0)
-		ssid->passphrase = NULL;
-        else
+	if (passphrase == NULL || strlen(passphrase) == 0) {
+
+		/* Use agent provided passphrase as a fallback */
+		agent_passphrase = connman_network_get_string(network,
+						"WiFi.AgentPassphrase");
+
+		if (agent_passphrase == NULL || strlen(agent_passphrase) == 0)
+			ssid->passphrase = NULL;
+		else
+			ssid->passphrase = agent_passphrase;
+	} else
 		ssid->passphrase = passphrase;
 
 	ssid->eap = connman_network_get_string(network, "WiFi.EAP");
@@ -379,6 +716,12 @@ static void ssid_init(GSupplicantSSID *ssid, struct connman_network *network)
 						ssid->passphrase);
 	/* We must have an identity for both PEAP and TLS */
 	ssid->identity = connman_network_get_string(network, "WiFi.Identity");
+
+	/* Use agent provided identity as a fallback */
+	if (ssid->identity == NULL || strlen(ssid->identity) == 0)
+		ssid->identity = connman_network_get_string(network,
+							"WiFi.AgentIdentity");
+
 	ssid->ca_cert_path = connman_network_get_string(network,
 							"WiFi.CACertFile");
 	ssid->client_cert_path = connman_network_get_string(network,
@@ -392,6 +735,7 @@ static void ssid_init(GSupplicantSSID *ssid, struct connman_network *network)
 	ssid->use_wps = connman_network_get_bool(network, "WiFi.UseWPS");
 	ssid->pin_wps = connman_network_get_string(network, "WiFi.PinWPS");
 
+	ssid->bgscan = BGSCAN_DEFAULT;
 }
 
 static int network_connect(struct connman_network *network)
@@ -419,9 +763,10 @@ static int network_connect(struct connman_network *network)
 	ssid_init(ssid, network);
 
 	if (wifi->disconnecting == TRUE)
-		wifi->pending_network = connman_network_ref(network);
+		wifi->pending_network = network;
 	else {
-		wifi->network = connman_network_ref(network);
+		wifi->network = network;
+		wifi->retries = 0;
 
 		return g_supplicant_interface_connect(interface, ssid,
 						connect_callback, network);
@@ -445,8 +790,6 @@ static void disconnect_callback(int result, GSupplicantInterface *interface,
 		 */
 		if (result < 0)
 			connman_network_set_connected(wifi->network, FALSE);
-
-		connman_network_unref(wifi->network);
 	}
 
 	wifi->network = NULL;
@@ -455,7 +798,6 @@ static void disconnect_callback(int result, GSupplicantInterface *interface,
 
 	if (wifi->pending_network != NULL) {
 		network_connect(wifi->pending_network);
-		connman_network_unref(wifi->pending_network);
 		wifi->pending_network = NULL;
 	}
 
@@ -526,8 +868,6 @@ static void interface_added(GSupplicantInterface *interface)
 
 	if (wifi->tethering == TRUE)
 		return;
-	
-	wifi_scan(wifi->device);
 }
 
 static connman_bool_t is_idle(struct wifi_data *wifi)
@@ -620,6 +960,23 @@ static connman_bool_t handle_wps_completion(GSupplicantInterface *interface,
 	return TRUE;
 }
 
+static connman_bool_t handle_4way_handshake_failure(GSupplicantInterface *interface,
+					struct connman_network *network,
+					struct wifi_data *wifi)
+{
+	if (wifi->state != G_SUPPLICANT_STATE_4WAY_HANDSHAKE)
+		return FALSE;
+
+	wifi->retries++;
+
+	if (wifi->retries < MAXIMUM_RETRIES)
+		return TRUE;
+
+	connman_network_set_error(network, CONNMAN_NETWORK_ERROR_INVALID_KEY);
+
+	return FALSE;
+}
+
 static void interface_state(GSupplicantInterface *interface)
 {
 	struct connman_network *network;
@@ -675,6 +1032,21 @@ static void interface_state(GSupplicantInterface *interface)
 
 		if (is_idle(wifi))
 			break;
+
+		/* If previous state was 4way-handshake, then
+		 * it's either: psk was incorrect and thus we retry
+		 * or if we reach the maximum retries we declare the
+		 * psk as wrong */
+		if (handle_4way_handshake_failure(interface,
+						network, wifi) == TRUE)
+			break;
+
+		/* We disable the selected network, if not then
+		 * wpa_supplicant will loop retrying */
+		if (g_supplicant_interface_enable_selected_network(interface,
+						FALSE) != 0)
+			DBG("Could not disables selected network");
+
 		connman_network_set_associating(network, FALSE);
 		connman_network_set_connected(network, FALSE);
 		break;
@@ -712,31 +1084,18 @@ static void interface_removed(GSupplicantInterface *interface)
 		return;
 	}
 
+	wifi->interface = NULL;
 	connman_device_set_powered(wifi->device, FALSE);
 }
 
 static void scan_started(GSupplicantInterface *interface)
 {
-	struct wifi_data *wifi;
-
 	DBG("");
-
-	wifi = g_supplicant_interface_get_data(interface);
-
-	if (wifi == NULL)
-		return;
 }
 
 static void scan_finished(GSupplicantInterface *interface)
 {
-	struct wifi_data *wifi;
-
 	DBG("");
-
-	wifi = g_supplicant_interface_get_data(interface);
-
-	if (wifi == NULL)
-		return;
 }
 
 static unsigned char calculate_strength(GSupplicantNetwork *supplicant_network)
@@ -755,7 +1114,7 @@ static void network_added(GSupplicantNetwork *supplicant_network)
 	struct connman_network *network;
 	GSupplicantInterface *interface;
 	struct wifi_data *wifi;
-	const char *name, *identifier, *mode, *security, *group;
+	const char *name, *identifier, *security, *group, *mode;
 	const unsigned char *ssid;
 	unsigned int ssid_len;
 	connman_bool_t wps;
@@ -766,10 +1125,10 @@ static void network_added(GSupplicantNetwork *supplicant_network)
 	wifi = g_supplicant_interface_get_data(interface);
 	name = g_supplicant_network_get_name(supplicant_network);
 	identifier = g_supplicant_network_get_identifier(supplicant_network);
-	mode = g_supplicant_network_get_mode(supplicant_network);
 	security = g_supplicant_network_get_security(supplicant_network);
 	group = g_supplicant_network_get_identifier(supplicant_network);
 	wps = g_supplicant_network_get_wps(supplicant_network);
+	mode = g_supplicant_network_get_mode(supplicant_network);
 
 	if (wifi == NULL)
 		return;
@@ -790,6 +1149,8 @@ static void network_added(GSupplicantNetwork *supplicant_network)
 			connman_network_unref(network);
 			return;
 		}
+
+		wifi->networks = g_slist_append(wifi->networks, network);
 	}
 
 	if (name != NULL && name[0] != '\0')
@@ -797,16 +1158,31 @@ static void network_added(GSupplicantNetwork *supplicant_network)
 
 	connman_network_set_blob(network, "WiFi.SSID",
 						ssid, ssid_len);
-	connman_network_set_string(network, "WiFi.Mode", mode);
 	connman_network_set_string(network, "WiFi.Security", security);
 	connman_network_set_strength(network,
 				calculate_strength(supplicant_network));
 	connman_network_set_bool(network, "WiFi.WPS", wps);
 
+	connman_network_set_frequency(network,
+			g_supplicant_network_get_frequency(supplicant_network));
+
 	connman_network_set_available(network, TRUE);
+	connman_network_set_string(network, "WiFi.Mode", mode);
 
 	if (ssid != NULL)
 		connman_network_set_group(network, group);
+
+	if (wifi->hidden != NULL) {
+		if (wifi->hidden->ssid_len == ssid_len &&
+				memcmp(wifi->hidden->ssid, ssid,
+						ssid_len) == 0) {
+			connman_network_connect_hidden(network,
+					wifi->hidden->identity,
+					wifi->hidden->passphrase);
+			hidden_free(wifi->hidden);
+			wifi->hidden = NULL;
+		}
+	}
 }
 
 static void network_removed(GSupplicantNetwork *network)
@@ -814,6 +1190,7 @@ static void network_removed(GSupplicantNetwork *network)
 	GSupplicantInterface *interface;
 	struct wifi_data *wifi;
 	const char *name, *identifier;
+	struct connman_network *connman_network;
 
 	interface = g_supplicant_network_get_interface(network);
 	wifi = g_supplicant_interface_get_data(interface);
@@ -822,8 +1199,45 @@ static void network_removed(GSupplicantNetwork *network)
 
 	DBG("name %s", name);
 
-	if (wifi != NULL)
-		connman_device_remove_network(wifi->device, identifier);
+	if (wifi == NULL)
+		return;
+
+	connman_network = connman_device_get_network(wifi->device, identifier);
+	if (connman_network == NULL)
+		return;
+
+	wifi->networks = g_slist_remove(wifi->networks, connman_network);
+
+	connman_device_remove_network(wifi->device, connman_network);
+	connman_network_unref(connman_network);
+}
+
+static void network_changed(GSupplicantNetwork *network, const char *property)
+{
+	GSupplicantInterface *interface;
+	struct wifi_data *wifi;
+	const char *name, *identifier;
+	struct connman_network *connman_network;
+
+	interface = g_supplicant_network_get_interface(network);
+	wifi = g_supplicant_interface_get_data(interface);
+	identifier = g_supplicant_network_get_identifier(network);
+	name = g_supplicant_network_get_name(network);
+
+	DBG("name %s", name);
+
+	if (wifi == NULL)
+		return;
+
+	connman_network = connman_device_get_network(wifi->device, identifier);
+	if (connman_network == NULL)
+		return;
+
+	if (g_str_equal(property, "Signal") == TRUE) {
+	       connman_network_set_strength(connman_network,
+					calculate_strength(network));
+	       connman_network_update(connman_network);
+	}
 }
 
 static void debug(const char *str)
@@ -842,6 +1256,7 @@ static const GSupplicantCallbacks callbacks = {
 	.scan_finished		= scan_finished,
 	.network_added		= network_added,
 	.network_removed	= network_removed,
+	.network_changed	= network_changed,
 	.debug			= debug,
 };
 
@@ -883,11 +1298,11 @@ static GSupplicantSSID *ssid_ap_init(const char *ssid, const char *passphrase)
 		ap->security = G_SUPPLICANT_SECURITY_NONE;
 		ap->passphrase = NULL;
 	} else {
-               ap->security = G_SUPPLICANT_SECURITY_PSK;
+	       ap->security = G_SUPPLICANT_SECURITY_PSK;
 	       ap->protocol = G_SUPPLICANT_PROTO_RSN;
 	       ap->pairwise_cipher = G_SUPPLICANT_PAIRWISE_CCMP;
 	       ap->group_cipher = G_SUPPLICANT_GROUP_CCMP;
-               ap->passphrase = passphrase;
+	       ap->passphrase = passphrase;
 	}
 
 	return ap;
@@ -916,7 +1331,6 @@ static void ap_create_callback(int result,
 					void *user_data)
 {
 	struct wifi_tethering_info *info = user_data;
-	struct connman_technology *technology;
 
 	DBG("result %d ifname %s", result,
 				g_supplicant_interface_get_ifname(interface));
@@ -932,7 +1346,6 @@ static void ap_create_callback(int result,
 	}
 
 	info->wifi->interface = interface;
-	technology = info->technology;
 	g_supplicant_interface_set_data(interface, info->wifi);
 
 	if (g_supplicant_interface_set_apscan(interface, 2) < 0)
