@@ -306,8 +306,15 @@ int __connman_device_disable(struct connman_device *device)
 
 	clear_scan_trigger(device);
 
-	if (device->network)
-		connman_network_set_connected(device->network, FALSE);
+	if (device->network) {
+		struct connman_service *service =
+			__connman_service_lookup_from_network(device->network);
+
+		if (service != NULL)
+			__connman_service_disconnect(service);
+		else
+			connman_network_set_connected(device->network, FALSE);
+	}
 
 	err = device->driver->disable(device);
 	if (err == 0) {
@@ -532,9 +539,11 @@ struct connman_device *connman_device_create(const char *node,
  *
  * Increase reference counter of device
  */
-struct connman_device *connman_device_ref(struct connman_device *device)
+struct connman_device *connman_device_ref_debug(struct connman_device *device,
+				const char *file, int line, const char *caller)
 {
-	DBG("%p", device);
+	DBG("%p ref %d by %s:%d:%s()", device, device->refcount + 1,
+		file, line, caller);
 
 	__sync_fetch_and_add(&device->refcount, 1);
 
@@ -547,8 +556,12 @@ struct connman_device *connman_device_ref(struct connman_device *device)
  *
  * Decrease reference counter of device
  */
-void connman_device_unref(struct connman_device *device)
+void connman_device_unref_debug(struct connman_device *device,
+				const char *file, int line, const char *caller)
 {
+	DBG("%p ref %d by %s:%d:%s()", device, device->refcount - 1,
+		file, line, caller);
+
 	if (__sync_fetch_and_sub(&device->refcount, 1) != 1)
 		return;
 
@@ -824,6 +837,8 @@ int connman_device_set_scanning(struct connman_device *device,
 	device->scanning = scanning;
 
 	if (scanning == TRUE) {
+		__connman_technology_scan_started(device);
+
 		reset_scan_trigger(device);
 
 		g_hash_table_foreach(device->networks,
@@ -834,6 +849,8 @@ int connman_device_set_scanning(struct connman_device *device,
 
 	__connman_device_cleanup_networks(device);
 
+	__connman_technology_scan_stopped(device);
+
 	__connman_service_auto_connect();
 
 	return 0;
@@ -1146,6 +1163,8 @@ struct connman_device *__connman_device_find_device(
 
 int __connman_device_request_scan(enum connman_service_type type)
 {
+	connman_bool_t success = FALSE;
+	int last_err = -ENOSYS;
 	GSList *list;
 	int err;
 
@@ -1158,7 +1177,7 @@ int __connman_device_request_scan(enum connman_service_type type)
 	case CONNMAN_SERVICE_TYPE_GPS:
 	case CONNMAN_SERVICE_TYPE_VPN:
 	case CONNMAN_SERVICE_TYPE_GADGET:
-		return 0;
+		return -EOPNOTSUPP;
 	case CONNMAN_SERVICE_TYPE_WIFI:
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 		break;
@@ -1175,14 +1194,35 @@ int __connman_device_request_scan(enum connman_service_type type)
 		}
 
 		err = device_scan(device);
-		if (err < 0 && err != -EINPROGRESS) {
-			DBG("err %d", err);
-			/* XXX maybe only a continue? */
-			return err;
+		if (err == 0 || err == -EALREADY || err == -EINPROGRESS) {
+			success = TRUE;
+		} else {
+			last_err = err;
+			DBG("device %p err %d", device, err);
 		}
 	}
 
-	return 0;
+	if (success == TRUE)
+		return 0;
+
+	return last_err;
+}
+
+int __connman_device_request_hidden_scan(struct connman_device *device,
+				const char *ssid, unsigned int ssid_len,
+				const char *identity, const char *passphrase)
+{
+	DBG("device %p", device);
+
+	if (device == NULL || device->driver == NULL ||
+			device->driver->scan_hidden == NULL)
+		return -EINVAL;
+
+	if (device->scanning == TRUE)
+		return -EALREADY;
+
+	return device->driver->scan_hidden(device, ssid, ssid_len,
+					identity, passphrase);
 }
 
 connman_bool_t __connman_device_isfiltered(const char *devname)
