@@ -23,6 +23,7 @@
 #include <config.h>
 #endif
 
+#include <errno.h>
 #include <stdio.h>
 #include <net/if.h>
 #include <net/if_arp.h>
@@ -39,7 +40,7 @@
 #include "connman.h"
 
 struct connman_ipconfig {
-	gint refcount;
+	int refcount;
 	int index;
 	enum connman_ipconfig_type type;
 
@@ -53,6 +54,7 @@ struct connman_ipconfig {
 	struct connman_ipaddress *system;
 
 	int ipv6_privacy_config;
+	char *last_dhcp_address;
 };
 
 struct connman_ipdevice {
@@ -450,6 +452,68 @@ static void set_ipv6_privacy(gchar *ifname, int value)
 	fclose(f);
 }
 
+static int get_rp_filter()
+{
+	FILE *f;
+	int value = -EINVAL, tmp;
+
+	f = fopen("/proc/sys/net/ipv4/conf/all/rp_filter", "r");
+
+	if (f != NULL) {
+		if (fscanf(f, "%d", &tmp) == 1)
+			value = tmp;
+		fclose(f);
+	}
+
+	return value;
+}
+
+static void set_rp_filter(int value)
+{
+	FILE *f;
+
+	f = fopen("/proc/sys/net/ipv4/conf/all/rp_filter", "r+");
+
+	if (f == NULL)
+		return;
+
+	fprintf(f, "%d", value);
+
+	fclose(f);
+}
+
+int __connman_ipconfig_set_rp_filter()
+{
+	int value;
+
+	value = get_rp_filter();
+
+	if (value < 0)
+		return value;
+
+	set_rp_filter(2);
+
+	connman_info("rp_filter set to 2 (loose mode routing), "
+			"old value was %d", value);
+
+	return value;
+}
+
+void __connman_ipconfig_unset_rp_filter(int old_value)
+{
+	set_rp_filter(old_value);
+
+	connman_info("rp_filter restored to %d", old_value);
+}
+
+gboolean __connman_ipconfig_ipv6_privacy_enabled(struct connman_ipconfig *ipconfig)
+{
+	if (ipconfig == NULL)
+		return FALSE;
+
+	return ipconfig->ipv6_privacy_config == 0 ? FALSE : TRUE;
+}
+
 static void free_ipdevice(gpointer data)
 {
 	struct connman_ipdevice *ipdevice = data;
@@ -458,12 +522,12 @@ static void free_ipdevice(gpointer data)
 							ipdevice->index);
 
 	if (ipdevice->config_ipv4 != NULL) {
-		connman_ipconfig_unref(ipdevice->config_ipv4);
+		__connman_ipconfig_unref(ipdevice->config_ipv4);
 		ipdevice->config_ipv4 = NULL;
 	}
 
 	if (ipdevice->config_ipv6 != NULL) {
-		connman_ipconfig_unref(ipdevice->config_ipv6);
+		__connman_ipconfig_unref(ipdevice->config_ipv6);
 		ipdevice->config_ipv6 = NULL;
 	}
 
@@ -519,9 +583,9 @@ static void update_stats(struct connman_ipdevice *ipdevice,
 		return;
 
 	if (ipdevice->config_ipv4)
-		service = connman_ipconfig_get_data(ipdevice->config_ipv4);
+		service = __connman_ipconfig_get_data(ipdevice->config_ipv4);
 	else if (ipdevice->config_ipv6)
-		service = connman_ipconfig_get_data(ipdevice->config_ipv6);
+		service = __connman_ipconfig_get_data(ipdevice->config_ipv6);
 	else
 		return;
 
@@ -707,6 +771,7 @@ void __connman_ipconfig_newaddr(int index, int family, const char *label,
 {
 	struct connman_ipdevice *ipdevice;
 	struct connman_ipaddress *ipaddress;
+	enum connman_ipconfig_type type;
 	GList *list;
 
 	DBG("index %d", index);
@@ -728,12 +793,22 @@ void __connman_ipconfig_newaddr(int index, int family, const char *label,
 		return;
 	}
 
+	if (family == AF_INET)
+		type = CONNMAN_IPCONFIG_TYPE_IPV4;
+	else if (family == AF_INET6)
+		type = CONNMAN_IPCONFIG_TYPE_IPV6;
+	else
+		return;
+
 	ipdevice->address_list = g_slist_append(ipdevice->address_list,
 								ipaddress);
 
 	connman_info("%s {add} address %s/%u label %s family %d",
 		ipdevice->ifname, address, prefixlen, label, family);
 
+	if (type == CONNMAN_IPCONFIG_TYPE_IPV4)
+		__connman_ippool_newaddr(index, address, prefixlen);
+
 	if (ipdevice->config_ipv4 != NULL && family == AF_INET)
 		connman_ipaddress_copy(ipdevice->config_ipv4->system,
 					ipaddress);
@@ -754,6 +829,9 @@ void __connman_ipconfig_newaddr(int index, int family, const char *label,
 		if (index != ipconfig->index)
 			continue;
 
+		if (type != ipconfig->type)
+			continue;
+
 		if (ipconfig->ops == NULL)
 			continue;
 
@@ -767,6 +845,7 @@ void __connman_ipconfig_deladdr(int index, int family, const char *label,
 {
 	struct connman_ipdevice *ipdevice;
 	struct connman_ipaddress *ipaddress;
+	enum connman_ipconfig_type type;
 	GList *list;
 
 	DBG("index %d", index);
@@ -779,9 +858,19 @@ void __connman_ipconfig_deladdr(int index, int family, const char *label,
 	if (ipaddress == NULL)
 		return;
 
+	if (family == AF_INET)
+		type = CONNMAN_IPCONFIG_TYPE_IPV4;
+	else if (family == AF_INET6)
+		type = CONNMAN_IPCONFIG_TYPE_IPV6;
+	else
+		return;
+
 	ipdevice->address_list = g_slist_remove(ipdevice->address_list,
 								ipaddress);
 
+	if (type == CONNMAN_IPCONFIG_TYPE_IPV4)
+		__connman_ippool_deladdr(index, address, prefixlen);
+
 	connman_ipaddress_clear(ipaddress);
 	g_free(ipaddress);
 
@@ -801,6 +890,9 @@ void __connman_ipconfig_deladdr(int index, int family, const char *label,
 		if (index != ipconfig->index)
 			continue;
 
+		if (type != ipconfig->type)
+			continue;
+
 		if (ipconfig->ops == NULL)
 			continue;
 
@@ -822,10 +914,11 @@ void __connman_ipconfig_newroute(int index, int family, unsigned char scope,
 
 	if (scope == 0 && (g_strcmp0(dst, "0.0.0.0") == 0 ||
 						g_strcmp0(dst, "::") == 0)) {
-		GSList *list;
 		GList *config_list;
+		enum connman_ipconfig_type type;
 
 		if (family == AF_INET6) {
+			type = CONNMAN_IPCONFIG_TYPE_IPV6;
 			g_free(ipdevice->ipv6_gateway);
 			ipdevice->ipv6_gateway = g_strdup(gateway);
 
@@ -836,6 +929,7 @@ void __connman_ipconfig_newroute(int index, int family, unsigned char scope,
 					g_strdup(gateway);
 			}
 		} else if (family == AF_INET) {
+			type = CONNMAN_IPCONFIG_TYPE_IPV4;
 			g_free(ipdevice->ipv4_gateway);
 			ipdevice->ipv4_gateway = g_strdup(gateway);
 
@@ -848,13 +942,6 @@ void __connman_ipconfig_newroute(int index, int family, unsigned char scope,
 		} else
 			return;
 
-		for (list = ipdevice->address_list; list; list = list->next) {
-			struct connman_ipaddress *ipaddress = list->data;
-
-			g_free(ipaddress->gateway);
-			ipaddress->gateway = g_strdup(gateway);
-		}
-
 		for (config_list = g_list_first(ipconfig_list); config_list;
 					config_list = g_list_next(config_list)) {
 			struct connman_ipconfig *ipconfig = config_list->data;
@@ -862,11 +949,14 @@ void __connman_ipconfig_newroute(int index, int family, unsigned char scope,
 			if (index != ipconfig->index)
 				continue;
 
+			if (type != ipconfig->type)
+				continue;
+
 			if (ipconfig->ops == NULL)
 				continue;
 
-			if (ipconfig->ops->ip_bound)
-				ipconfig->ops->ip_bound(ipconfig);
+			if (ipconfig->ops->route_set)
+				ipconfig->ops->route_set(ipconfig);
 		}
 	}
 
@@ -888,10 +978,11 @@ void __connman_ipconfig_delroute(int index, int family, unsigned char scope,
 
 	if (scope == 0 && (g_strcmp0(dst, "0.0.0.0") == 0 ||
 						g_strcmp0(dst, "::") == 0)) {
-		GSList *list;
 		GList *config_list;
+		enum connman_ipconfig_type type;
 
 		if (family == AF_INET6) {
+			type = CONNMAN_IPCONFIG_TYPE_IPV6;
 			g_free(ipdevice->ipv6_gateway);
 			ipdevice->ipv6_gateway = NULL;
 
@@ -901,6 +992,7 @@ void __connman_ipconfig_delroute(int index, int family, unsigned char scope,
 				ipdevice->config_ipv6->system->gateway = NULL;
 			}
 		} else if (family == AF_INET) {
+			type = CONNMAN_IPCONFIG_TYPE_IPV4;
 			g_free(ipdevice->ipv4_gateway);
 			ipdevice->ipv4_gateway = NULL;
 
@@ -912,13 +1004,6 @@ void __connman_ipconfig_delroute(int index, int family, unsigned char scope,
 		} else
 			return;
 
-		for (list = ipdevice->address_list; list; list = list->next) {
-			struct connman_ipaddress *ipaddress = list->data;
-
-			g_free(ipaddress->gateway);
-			ipaddress->gateway = NULL;
-		}
-
 		for (config_list = g_list_first(ipconfig_list); config_list;
 					config_list = g_list_next(config_list)) {
 			struct connman_ipconfig *ipconfig = config_list->data;
@@ -926,11 +1011,14 @@ void __connman_ipconfig_delroute(int index, int family, unsigned char scope,
 			if (index != ipconfig->index)
 				continue;
 
+			if (type != ipconfig->type)
+				continue;
+
 			if (ipconfig->ops == NULL)
 				continue;
 
-			if (ipconfig->ops->ip_release)
-				ipconfig->ops->ip_release(ipconfig);
+			if (ipconfig->ops->route_unset)
+				ipconfig->ops->route_unset(ipconfig);
 		}
 	}
 
@@ -1097,17 +1185,17 @@ int __connman_ipconfig_gateway_add(struct connman_ipconfig *ipconfig)
 	if (service == NULL)
 		return -EINVAL;
 
-	__connman_connection_gateway_remove(service);
+	__connman_connection_gateway_remove(service, ipconfig->type);
 
-	if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV6) {
-		return __connman_connection_gateway_add(service, NULL,
-						ipconfig->address->gateway,
-							ipconfig->address->peer);
-	} else if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV4) {
+	DBG("type %d gw %s peer %s", ipconfig->type,
+		ipconfig->address->gateway, ipconfig->address->peer);
+
+	if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV6 ||
+				ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV4)
 		return __connman_connection_gateway_add(service,
 						ipconfig->address->gateway,
-							NULL, ipconfig->address->peer);
-	}
+						ipconfig->type,
+						ipconfig->address->peer);
 
 	return 0;
 }
@@ -1120,7 +1208,7 @@ void __connman_ipconfig_gateway_remove(struct connman_ipconfig *ipconfig)
 
 	service = __connman_service_lookup_from_index(ipconfig->index);
 	if (service != NULL)
-		__connman_connection_gateway_remove(service);
+		__connman_connection_gateway_remove(service, ipconfig->type);
 }
 
 unsigned char __connman_ipconfig_get_prefixlen(struct connman_ipconfig *ipconfig)
@@ -1176,7 +1264,7 @@ static struct connman_ipconfig *create_ipv6config(int index)
  *
  * Returns: a newly-allocated #connman_ipconfig structure
  */
-struct connman_ipconfig *connman_ipconfig_create(int index,
+struct connman_ipconfig *__connman_ipconfig_create(int index,
 					enum connman_ipconfig_type type)
 {
 	struct connman_ipconfig *ipconfig;
@@ -1215,12 +1303,14 @@ struct connman_ipconfig *connman_ipconfig_create(int index,
  *
  * Increase reference counter of ipconfig
  */
-struct connman_ipconfig *connman_ipconfig_ref(struct connman_ipconfig *ipconfig)
+struct connman_ipconfig *
+__connman_ipconfig_ref_debug(struct connman_ipconfig *ipconfig,
+				const char *file, int line, const char *caller)
 {
-	DBG("ipconfig %p refcount %d", ipconfig,
-				g_atomic_int_get(&ipconfig->refcount) + 1);
+	DBG("%p ref %d by %s:%d:%s()", ipconfig, ipconfig->refcount + 1,
+		file, line, caller);
 
-	g_atomic_int_inc(&ipconfig->refcount);
+	__sync_fetch_and_add(&ipconfig->refcount, 1);
 
 	return ipconfig;
 }
@@ -1231,28 +1321,32 @@ struct connman_ipconfig *connman_ipconfig_ref(struct connman_ipconfig *ipconfig)
  *
  * Decrease reference counter of ipconfig
  */
-void connman_ipconfig_unref(struct connman_ipconfig *ipconfig)
+void __connman_ipconfig_unref_debug(struct connman_ipconfig *ipconfig,
+				const char *file, int line, const char *caller)
 {
 	if (ipconfig == NULL)
 		return;
 
-	DBG("ipconfig %p refcount %d", ipconfig,
-			g_atomic_int_get(&ipconfig->refcount) - 1);
+	DBG("%p ref %d by %s:%d:%s()", ipconfig, ipconfig->refcount - 1,
+		file, line, caller);
 
-	if (g_atomic_int_dec_and_test(&ipconfig->refcount) == TRUE) {
-		__connman_ipconfig_disable(ipconfig);
+	if (__sync_fetch_and_sub(&ipconfig->refcount, 1) != 1)
+		return;
 
-		connman_ipconfig_set_ops(ipconfig, NULL);
+	if (__connman_ipconfig_disable(ipconfig) < 0)
+		ipconfig_list = g_list_remove(ipconfig_list, ipconfig);
 
-		if (ipconfig->origin != NULL) {
-			connman_ipconfig_unref(ipconfig->origin);
-			ipconfig->origin = NULL;
-		}
+	__connman_ipconfig_set_ops(ipconfig, NULL);
 
-		connman_ipaddress_free(ipconfig->system);
-		connman_ipaddress_free(ipconfig->address);
-		g_free(ipconfig);
+	if (ipconfig->origin != NULL && ipconfig->origin != ipconfig) {
+		__connman_ipconfig_unref(ipconfig->origin);
+		ipconfig->origin = NULL;
 	}
+
+	connman_ipaddress_free(ipconfig->system);
+	connman_ipaddress_free(ipconfig->address);
+	g_free(ipconfig->last_dhcp_address);
+	g_free(ipconfig);
 }
 
 /**
@@ -1261,7 +1355,7 @@ void connman_ipconfig_unref(struct connman_ipconfig *ipconfig)
  *
  * Get private data pointer
  */
-void *connman_ipconfig_get_data(struct connman_ipconfig *ipconfig)
+void *__connman_ipconfig_get_data(struct connman_ipconfig *ipconfig)
 {
 	if (ipconfig == NULL)
 		return NULL;
@@ -1276,7 +1370,7 @@ void *connman_ipconfig_get_data(struct connman_ipconfig *ipconfig)
  *
  * Set private data pointer
  */
-void connman_ipconfig_set_data(struct connman_ipconfig *ipconfig, void *data)
+void __connman_ipconfig_set_data(struct connman_ipconfig *ipconfig, void *data)
 {
 	ipconfig->ops_data = data;
 }
@@ -1287,7 +1381,7 @@ void connman_ipconfig_set_data(struct connman_ipconfig *ipconfig, void *data)
  *
  * Get interface index
  */
-int connman_ipconfig_get_index(struct connman_ipconfig *ipconfig)
+int __connman_ipconfig_get_index(struct connman_ipconfig *ipconfig)
 {
 	if (ipconfig == NULL)
 		return -1;
@@ -1304,7 +1398,7 @@ int connman_ipconfig_get_index(struct connman_ipconfig *ipconfig)
  *
  * Get interface name
  */
-const char *connman_ipconfig_get_ifname(struct connman_ipconfig *ipconfig)
+const char *__connman_ipconfig_get_ifname(struct connman_ipconfig *ipconfig)
 {
 	struct connman_ipdevice *ipdevice;
 
@@ -1329,7 +1423,7 @@ const char *connman_ipconfig_get_ifname(struct connman_ipconfig *ipconfig)
  *
  * Set the operation callbacks
  */
-void connman_ipconfig_set_ops(struct connman_ipconfig *ipconfig,
+void __connman_ipconfig_set_ops(struct connman_ipconfig *ipconfig,
 				const struct connman_ipconfig_ops *ops)
 {
 	ipconfig->ops = ops;
@@ -1342,7 +1436,7 @@ void connman_ipconfig_set_ops(struct connman_ipconfig *ipconfig,
  *
  * Set the configuration method
  */
-int connman_ipconfig_set_method(struct connman_ipconfig *ipconfig,
+int __connman_ipconfig_set_method(struct connman_ipconfig *ipconfig,
 					enum connman_ipconfig_method method)
 {
 	ipconfig->method = method;
@@ -1365,8 +1459,8 @@ int __connman_ipconfig_address_add(struct connman_ipconfig *ipconfig)
 	switch (ipconfig->method) {
 	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
 	case CONNMAN_IPCONFIG_METHOD_OFF:
-	case CONNMAN_IPCONFIG_METHOD_AUTO:
 		break;
+	case CONNMAN_IPCONFIG_METHOD_AUTO:
 	case CONNMAN_IPCONFIG_METHOD_FIXED:
 	case CONNMAN_IPCONFIG_METHOD_DHCP:
 	case CONNMAN_IPCONFIG_METHOD_MANUAL:
@@ -1395,8 +1489,36 @@ int __connman_ipconfig_address_remove(struct connman_ipconfig *ipconfig)
 	switch (ipconfig->method) {
 	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
 	case CONNMAN_IPCONFIG_METHOD_OFF:
-	case CONNMAN_IPCONFIG_METHOD_AUTO:
 		break;
+	case CONNMAN_IPCONFIG_METHOD_AUTO:
+	case CONNMAN_IPCONFIG_METHOD_FIXED:
+	case CONNMAN_IPCONFIG_METHOD_DHCP:
+	case CONNMAN_IPCONFIG_METHOD_MANUAL:
+		err = __connman_ipconfig_address_unset(ipconfig);
+		connman_ipaddress_clear(ipconfig->address);
+
+		return err;
+	}
+
+	return 0;
+}
+
+int __connman_ipconfig_address_unset(struct connman_ipconfig *ipconfig)
+{
+	int err;
+
+	DBG("");
+
+	if (ipconfig == NULL)
+		return 0;
+
+	DBG("method %d", ipconfig->method);
+
+	switch (ipconfig->method) {
+	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
+	case CONNMAN_IPCONFIG_METHOD_OFF:
+		break;
+	case CONNMAN_IPCONFIG_METHOD_AUTO:
 	case CONNMAN_IPCONFIG_METHOD_FIXED:
 	case CONNMAN_IPCONFIG_METHOD_DHCP:
 	case CONNMAN_IPCONFIG_METHOD_MANUAL:
@@ -1411,8 +1533,6 @@ int __connman_ipconfig_address_remove(struct connman_ipconfig *ipconfig)
 		else
 			err = -EINVAL;
 
-		connman_ipaddress_clear(ipconfig->address);
-
 		return err;
 	}
 
@@ -1457,6 +1577,24 @@ const char *__connman_ipconfig_get_proxy_autoconfig(struct connman_ipconfig *ipc
 	return ipdevice->pac;
 }
 
+void __connman_ipconfig_set_dhcp_address(struct connman_ipconfig *ipconfig,
+					const char *address)
+{
+	if (ipconfig == NULL)
+		return;
+
+	g_free(ipconfig->last_dhcp_address);
+	ipconfig->last_dhcp_address = g_strdup(address);
+}
+
+char *__connman_ipconfig_get_dhcp_address(struct connman_ipconfig *ipconfig)
+{
+	if (ipconfig == NULL)
+		return NULL;
+
+	return ipconfig->last_dhcp_address;
+}
+
 static void disable_ipv6(struct connman_ipconfig *ipconfig)
 {
 	struct connman_ipdevice *ipdevice;
@@ -1489,6 +1627,14 @@ static void enable_ipv6(struct connman_ipconfig *ipconfig)
 	set_ipv6_state(ipdevice->ifname, TRUE);
 }
 
+void __connman_ipconfig_enable_ipv6(struct connman_ipconfig *ipconfig)
+{
+	if (ipconfig == NULL || ipconfig->type != CONNMAN_IPCONFIG_TYPE_IPV6)
+		return;
+
+	enable_ipv6(ipconfig);
+}
+
 void __connman_ipconfig_disable_ipv6(struct connman_ipconfig *ipconfig)
 {
 	if (ipconfig == NULL || ipconfig->type != CONNMAN_IPCONFIG_TYPE_IPV6)
@@ -1533,7 +1679,7 @@ int __connman_ipconfig_enable(struct connman_ipconfig *ipconfig)
 
 		connman_ipaddress_clear(ipdevice->config_ipv4->system);
 
-		connman_ipconfig_unref(ipdevice->config_ipv4);
+		__connman_ipconfig_unref(ipdevice->config_ipv4);
 	}
 
 	if (type == CONNMAN_IPCONFIG_TYPE_IPV6 &&
@@ -1543,13 +1689,13 @@ int __connman_ipconfig_enable(struct connman_ipconfig *ipconfig)
 
 		connman_ipaddress_clear(ipdevice->config_ipv6->system);
 
-		connman_ipconfig_unref(ipdevice->config_ipv6);
+		__connman_ipconfig_unref(ipdevice->config_ipv6);
 	}
 
 	if (type == CONNMAN_IPCONFIG_TYPE_IPV4)
-		ipdevice->config_ipv4 = connman_ipconfig_ref(ipconfig);
+		ipdevice->config_ipv4 = __connman_ipconfig_ref(ipconfig);
 	else if (type == CONNMAN_IPCONFIG_TYPE_IPV6)
-		ipdevice->config_ipv6 = connman_ipconfig_ref(ipconfig);
+		ipdevice->config_ipv6 = __connman_ipconfig_ref(ipconfig);
 
 	ipconfig_list = g_list_append(ipconfig_list, ipconfig);
 
@@ -1598,7 +1744,7 @@ int __connman_ipconfig_disable(struct connman_ipconfig *ipconfig)
 		ipconfig_list = g_list_remove(ipconfig_list, ipconfig);
 
 		connman_ipaddress_clear(ipdevice->config_ipv4->system);
-		connman_ipconfig_unref(ipdevice->config_ipv4);
+		__connman_ipconfig_unref(ipdevice->config_ipv4);
 		ipdevice->config_ipv4 = NULL;
 		return 0;
 	}
@@ -1611,7 +1757,7 @@ int __connman_ipconfig_disable(struct connman_ipconfig *ipconfig)
 			disable_ipv6(ipdevice->config_ipv6);
 
 		connman_ipaddress_clear(ipdevice->config_ipv6->system);
-		connman_ipconfig_unref(ipdevice->config_ipv6);
+		__connman_ipconfig_unref(ipdevice->config_ipv6);
 		ipdevice->config_ipv6 = NULL;
 		return 0;
 	}
@@ -1821,10 +1967,10 @@ void __connman_ipconfig_append_ipv4config(struct connman_ipconfig *ipconfig,
 	switch (ipconfig->method) {
 	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
 	case CONNMAN_IPCONFIG_METHOD_OFF:
-	case CONNMAN_IPCONFIG_METHOD_FIXED:
 	case CONNMAN_IPCONFIG_METHOD_DHCP:
 	case CONNMAN_IPCONFIG_METHOD_AUTO:
 		return;
+	case CONNMAN_IPCONFIG_METHOD_FIXED:
 	case CONNMAN_IPCONFIG_METHOD_MANUAL:
 		break;
 	}
@@ -1869,7 +2015,7 @@ int __connman_ipconfig_set_config(struct connman_ipconfig *ipconfig,
 	dbus_message_iter_recurse(array, &dict);
 
 	while (dbus_message_iter_get_arg_type(&dict) == DBUS_TYPE_DICT_ENTRY) {
-		DBusMessageIter entry;
+		DBusMessageIter entry, value;
 		const char *key;
 		int type;
 
@@ -1881,7 +2027,12 @@ int __connman_ipconfig_set_config(struct connman_ipconfig *ipconfig,
 		dbus_message_iter_get_basic(&entry, &key);
 		dbus_message_iter_next(&entry);
 
-		type = dbus_message_iter_get_arg_type(&entry);
+		if (dbus_message_iter_get_arg_type(&entry) != DBUS_TYPE_VARIANT)
+			return -EINVAL;
+
+		dbus_message_iter_recurse(&entry, &value);
+
+		type = dbus_message_iter_get_arg_type(&value);
 
 		if (g_str_equal(key, "Method") == TRUE) {
 			const char *str;
@@ -1889,41 +2040,41 @@ int __connman_ipconfig_set_config(struct connman_ipconfig *ipconfig,
 			if (type != DBUS_TYPE_STRING)
 				return -EINVAL;
 
-			dbus_message_iter_get_basic(&entry, &str);
+			dbus_message_iter_get_basic(&value, &str);
 			method = __connman_ipconfig_string2method(str);
 		} else if (g_str_equal(key, "Address") == TRUE) {
 			if (type != DBUS_TYPE_STRING)
 				return -EINVAL;
 
-			dbus_message_iter_get_basic(&entry, &address);
+			dbus_message_iter_get_basic(&value, &address);
 		} else if (g_str_equal(key, "PrefixLength") == TRUE) {
 			if (type != DBUS_TYPE_STRING)
 				return -EINVAL;
 
-			dbus_message_iter_get_basic(&entry,
+			dbus_message_iter_get_basic(&value,
 							&prefix_length_string);
 
 			prefix_length = atoi(prefix_length_string);
 			if (prefix_length < 0 || prefix_length > 128)
 				return -EINVAL;
-
 		} else if (g_str_equal(key, "Netmask") == TRUE) {
 			if (type != DBUS_TYPE_STRING)
 				return -EINVAL;
 
-			dbus_message_iter_get_basic(&entry, &netmask);
+			dbus_message_iter_get_basic(&value, &netmask);
 		} else if (g_str_equal(key, "Gateway") == TRUE) {
 			if (type != DBUS_TYPE_STRING)
 				return -EINVAL;
 
-			dbus_message_iter_get_basic(&entry, &gateway);
+			dbus_message_iter_get_basic(&value, &gateway);
 		} else if (g_str_equal(key, "Privacy") == TRUE) {
 			if (type != DBUS_TYPE_STRING)
 				return -EINVAL;
 
-			dbus_message_iter_get_basic(&entry, &privacy_string);
+			dbus_message_iter_get_basic(&value, &privacy_string);
 			privacy = string2privacy(privacy_string);
 		}
+
 		dbus_message_iter_next(&dict);
 	}
 
@@ -2011,16 +2162,24 @@ int __connman_ipconfig_load(struct connman_ipconfig *ipconfig,
 {
 	char *method;
 	char *key;
+	char *str;
 
 	DBG("ipconfig %p identifier %s", ipconfig, identifier);
 
 	key = g_strdup_printf("%smethod", prefix);
 	method = g_key_file_get_string(keyfile, identifier, key, NULL);
 	if (method == NULL) {
-		if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV4)
+		switch (ipconfig->type) {
+		case CONNMAN_IPCONFIG_TYPE_IPV4:
 			ipconfig->method = CONNMAN_IPCONFIG_METHOD_DHCP;
-		else
+			break;
+		case CONNMAN_IPCONFIG_TYPE_IPV6:
+			ipconfig->method = CONNMAN_IPCONFIG_METHOD_AUTO;
+			break;
+		case CONNMAN_IPCONFIG_TYPE_UNKNOWN:
 			ipconfig->method = CONNMAN_IPCONFIG_METHOD_OFF;
+			break;
+		}
 	} else
 		ipconfig->method = __connman_ipconfig_string2method(method);
 
@@ -2071,6 +2230,14 @@ int __connman_ipconfig_load(struct connman_ipconfig *ipconfig,
 				keyfile, identifier, key, NULL);
 	g_free(key);
 
+	key = g_strdup_printf("%sDHCP.LastAddress", prefix);
+	str = g_key_file_get_string(keyfile, identifier, key, NULL);
+	if (str != NULL) {
+		g_free(ipconfig->last_dhcp_address);
+		ipconfig->last_dhcp_address = str;
+	}
+	g_free(key);
+
 	return 0;
 }
 
@@ -2100,9 +2267,18 @@ int __connman_ipconfig_save(struct connman_ipconfig *ipconfig,
 	case CONNMAN_IPCONFIG_METHOD_FIXED:
 	case CONNMAN_IPCONFIG_METHOD_MANUAL:
 		break;
+	case CONNMAN_IPCONFIG_METHOD_DHCP:
+		key = g_strdup_printf("%sDHCP.LastAddress", prefix);
+		if (ipconfig->last_dhcp_address != NULL &&
+				strlen(ipconfig->last_dhcp_address) > 0)
+			g_key_file_set_string(keyfile, identifier, key,
+					ipconfig->last_dhcp_address);
+		else
+			g_key_file_remove_key(keyfile, identifier, key, NULL);
+		g_free(key);
+		/* fall through */
 	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
 	case CONNMAN_IPCONFIG_METHOD_OFF:
-	case CONNMAN_IPCONFIG_METHOD_DHCP:
 	case CONNMAN_IPCONFIG_METHOD_AUTO:
 		return 0;
 	}
