@@ -76,20 +76,14 @@ void __connman_provider_append_properties(struct connman_provider *provider,
 						 &provider->type);
 }
 
-static int connman_provider_load(struct connman_provider *provider)
+static int provider_load_from_keyfile(struct connman_provider *provider,
+		GKeyFile *keyfile)
 {
 	gsize idx = 0;
-	GKeyFile *keyfile;
 	gchar **settings;
 	gchar *key, *value;
 	gsize length;
 
-	DBG("provider %p", provider);
-
-	keyfile = __connman_storage_load_provider(provider->identifier);
-	if (keyfile == NULL)
-		return -ENOENT;
-
 	settings = g_key_file_get_keys(keyfile, provider->identifier, &length,
 				NULL);
 	if (settings == NULL) {
@@ -99,7 +93,6 @@ static int connman_provider_load(struct connman_provider *provider)
 
 	while (idx < length) {
 		key = settings[idx];
-		DBG("found key %s", key);
 		if (key != NULL) {
 			value = g_key_file_get_string(keyfile,
 						provider->identifier,
@@ -111,6 +104,21 @@ static int connman_provider_load(struct connman_provider *provider)
 	}
 	g_strfreev(settings);
 
+	return 0;
+}
+
+static int connman_provider_load(struct connman_provider *provider)
+{
+	GKeyFile *keyfile;
+
+	DBG("provider %p", provider);
+
+	keyfile = __connman_storage_load_provider(provider->identifier);
+	if (keyfile == NULL)
+		return -ENOENT;
+
+	provider_load_from_keyfile(provider, keyfile);
+
 	g_key_file_free(keyfile);
 	return 0;
 }
@@ -200,7 +208,6 @@ static void provider_remove(struct connman_provider *provider)
 
 static int provider_register(struct connman_provider *provider)
 {
-	connman_provider_load(provider);
 	return provider_probe(provider);
 }
 
@@ -209,9 +216,12 @@ static void provider_unregister(struct connman_provider *provider)
 	provider_remove(provider);
 }
 
-struct connman_provider *connman_provider_ref(struct connman_provider *provider)
+struct connman_provider *
+connman_provider_ref_debug(struct connman_provider *provider,
+			const char *file, int line, const char *caller)
 {
-	DBG("provider %p refcount %d", provider, provider->refcount + 1);
+	DBG("%p ref %d by %s:%d:%s()", provider, provider->refcount + 1,
+		file, line, caller);
 
 	__sync_fetch_and_add(&provider->refcount, 1);
 
@@ -231,9 +241,11 @@ static void provider_destruct(struct connman_provider *provider)
 	g_hash_table_destroy(provider->setting_strings);
 }
 
-void connman_provider_unref(struct connman_provider *provider)
+void connman_provider_unref_debug(struct connman_provider *provider,
+				const char *file, int line, const char *caller)
 {
-	DBG("provider %p refcount %d", provider, provider->refcount - 1);
+	DBG("%p ref %d by %s:%d:%s()", provider, provider->refcount - 1,
+		file, line, caller);
 
 	if (__sync_fetch_and_sub(&provider->refcount, 1) != 1)
 		return;
@@ -427,8 +439,6 @@ int connman_provider_indicate_error(struct connman_provider *provider,
 					enum connman_provider_error error)
 {
 	enum connman_service_error service_error;
-	const char *path;
-	int ret;
 
 	switch (error) {
 	case CONNMAN_PROVIDER_ERROR_LOGIN_FAILED:
@@ -445,23 +455,20 @@ int connman_provider_indicate_error(struct connman_provider *provider,
 		break;
 	}
 
-	ret = __connman_service_indicate_error(provider->vpn_service,
+	return __connman_service_indicate_error(provider->vpn_service,
 							service_error);
-	path = __connman_service_get_path(provider->vpn_service);
-	__connman_provider_remove(path);
-
-	return ret;
 }
 
 static void unregister_provider(gpointer data)
 {
 	struct connman_provider *provider = data;
-	struct connman_service *service = provider->vpn_service;
 
-	DBG("provider %p", provider);
+	DBG("provider %p service %p", provider, provider->vpn_service);
 
-	provider->vpn_service = NULL;
-	__connman_service_put(service);
+	if (provider->vpn_service != NULL) {
+		connman_service_unref(provider->vpn_service);
+		provider->vpn_service = NULL;
+	}
 
 	connman_provider_unref(provider);
 }
@@ -525,8 +532,6 @@ static struct connman_provider *connman_provider_get(const char *identifier)
 
 	g_hash_table_insert(provider_hash, provider->identifier, provider);
 
-	provider->name = g_strdup(identifier);
-
 	return provider;
 }
 
@@ -545,6 +550,97 @@ static void provider_dbus_ident(char *ident)
 	}
 }
 
+static struct connman_provider *provider_create_from_keyfile(GKeyFile *keyfile,
+		const char *ident)
+{
+	struct connman_provider *provider;
+
+	if (keyfile == NULL || ident == NULL)
+		return NULL;
+
+	provider = connman_provider_lookup(ident);
+	if (provider == NULL) {
+		provider = connman_provider_get(ident);
+		if (provider == NULL) {
+			DBG("can not create provider");
+			return NULL;
+		}
+
+		provider_load_from_keyfile(provider, keyfile);
+
+		if (provider->name == NULL || provider->host == NULL ||
+				provider->domain == NULL) {
+			DBG("cannot get name, host or domain");
+			connman_provider_unref(provider);
+			return NULL;
+		}
+
+		provider_register(provider);
+	}
+	return provider;
+}
+
+static int provider_create_service(struct connman_provider *provider)
+{
+	if (provider->vpn_service != NULL)
+		return -EALREADY;
+
+	provider->vpn_service =
+		__connman_service_create_from_provider(provider);
+
+	if (provider->vpn_service == NULL)
+		return -EOPNOTSUPP;
+
+	return 0;
+}
+
+static void provider_create_all_from_type(const char *provider_type)
+{
+	unsigned int i;
+	char **providers;
+	char *id, *type;
+	GKeyFile *keyfile;
+	struct connman_provider *provider;
+
+	DBG("provider type %s", provider_type);
+
+	providers = __connman_storage_get_providers();
+
+	for (i = 0; providers[i] != NULL; i+=1) {
+
+		if (strncmp(providers[i], "provider_", 9) != 0)
+			continue;
+
+		id = providers[i] + 9;
+		keyfile = __connman_storage_load_provider(id);
+
+		if (keyfile == NULL)
+			continue;
+
+		type = g_key_file_get_string(keyfile, id, "Type", NULL);
+
+		DBG("keyfile %p id %s type %s", keyfile, id, type);
+
+		if (strcmp(provider_type, type) != 0) {
+			g_free(type);
+			g_key_file_free(keyfile);
+			continue;
+		}
+
+		provider = provider_create_from_keyfile(keyfile, id);
+		if (provider != NULL) {
+			if (provider_create_service(provider) == -EOPNOTSUPP) {
+				DBG("could not create service");
+				connman_provider_unref(provider);
+			}
+		}
+
+		g_free(type);
+		g_key_file_free(keyfile);
+	}
+	g_strfreev(providers);
+}
+
 int __connman_provider_create_and_connect(DBusMessage *msg)
 {
 	struct connman_provider *provider;
@@ -607,11 +703,11 @@ int __connman_provider_create_and_connect(DBusMessage *msg)
 
 		provider->host = g_strdup(host);
 		provider->domain = g_strdup(domain);
-		g_free(provider->name);
 		provider->name = g_strdup(name);
 		provider->type = g_strdup(type);
 
-		provider_register(provider);
+		if (provider_register(provider) == 0)
+			connman_provider_load(provider);
 	}
 
 	dbus_message_iter_init(msg, &iter);
@@ -639,19 +735,19 @@ int __connman_provider_create_and_connect(DBusMessage *msg)
 
 	g_free(ident);
 
-	if (provider->vpn_service == NULL) {
-		provider->vpn_service =
-			__connman_service_create_from_provider(provider);
-		if (provider->vpn_service == NULL) {
-			err = -EOPNOTSUPP;
-			goto unref;
-		}
-
-		err = __connman_service_connect(provider->vpn_service);
-		if (err < 0 && err != -EINPROGRESS)
-			goto failed;
-	} else
+	err = provider_create_service(provider);
+	if (err == -EALREADY) {
 		DBG("provider already connected");
+	} else {
+		if (err == -EOPNOTSUPP) {
+			goto unref;
+		} else {
+			err = __connman_service_connect(provider->vpn_service);
+
+			if (err < 0 && err != -EINPROGRESS)
+				goto failed;
+		}
+	}
 
 	connman_provider_save(provider);
 	service_path = __connman_service_get_path(provider->vpn_service);
@@ -661,13 +757,13 @@ int __connman_provider_create_and_connect(DBusMessage *msg)
 	return 0;
 
 failed:
-	__connman_service_put(provider->vpn_service);
+	connman_service_unref(provider->vpn_service);
 	provider->vpn_service = NULL;
 
 unref:
 	DBG("can not connect, delete provider");
 
-	connman_provider_unref(provider);
+	g_hash_table_remove(provider_hash, provider->identifier);
 
 	return err;
 }
@@ -752,7 +848,7 @@ void connman_provider_set_index(struct connman_provider *provider, int index)
 		}
 	}
 
-	connman_ipconfig_set_method(ipconfig, CONNMAN_IPCONFIG_METHOD_FIXED);
+	__connman_ipconfig_set_method(ipconfig, CONNMAN_IPCONFIG_METHOD_FIXED);
 	__connman_ipconfig_set_index(ipconfig, index);
 
 
@@ -768,7 +864,7 @@ void connman_provider_set_index(struct connman_provider *provider, int index)
 		}
 	}
 
-	connman_ipconfig_set_method(ipconfig, CONNMAN_IPCONFIG_METHOD_OFF);
+	__connman_ipconfig_set_method(ipconfig, CONNMAN_IPCONFIG_METHOD_OFF);
 	__connman_ipconfig_set_index(ipconfig, index);
 
 done:
@@ -986,6 +1082,7 @@ int connman_provider_driver_register(struct connman_provider_driver *driver)
 
 	driver_list = g_slist_insert_sorted(driver_list, driver,
 							compare_priority);
+	provider_create_all_from_type(driver->name);
 	return 0;
 }
 
